#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_fulfil_client
----------------------------------

Tests for `fulfil_client` module.
"""
import pytest
from fulfil_client import Client, ClientError, ServerError


def test_find(client):
    IRModel = client.model('ir.model')
    ir_models = IRModel.find([])
    assert len(ir_models) > 0
    assert ir_models[0]['id']
    assert ir_models[0]['rec_name']


def test_search_read_all(client):
    IRModel = client.model('ir.model')
    total_records = IRModel.search_count([])
    ir_models = list(
        IRModel.search_read_all([], None, ['rec_name'], batch_size=50)
    )

    # the default batch size is 500 and the total records
    # being greater than that is an important part of the test.
    assert total_records > 500
    assert total_records == len(set([r['id'] for r in ir_models]))

    assert len(ir_models) == total_records
    assert len(ir_models) > 10
    assert ir_models[0]['id']
    assert ir_models[0]['rec_name']

    first_record = ir_models[0]

    # Offset and then fetch
    ir_models = list(
        IRModel.search_read_all(
            [], None, ['rec_name'], offset=10
        )
    )
    assert len(ir_models) == total_records - 10
    assert ir_models[0]['id'] != first_record['id']

    # Smaller batch size and offset
    ir_models = list(
        IRModel.search_read_all(
            [], None, ['rec_name'], batch_size=5, offset=10
        )
    )
    assert len(ir_models) == total_records - 10
    assert ir_models[0]['id'] != first_record['id']

    # Smaller batch size and limit
    ir_models = list(
        IRModel.search_read_all(
            [], None, ['rec_name'], batch_size=5, limit=10
        )
    )
    assert len(ir_models) == 10
    assert ir_models[0]['id'] == first_record['id']

    #  default batch size and limit
    ir_models = list(
        IRModel.search_read_all(
            [], None, ['rec_name'], limit=10
        )
    )
    assert len(ir_models) == 10
    assert ir_models[0]['id'] == first_record['id']

    # small batch size and limit and offset
    ir_models = list(
        IRModel.search_read_all(
            [], None, ['rec_name'],
            batch_size=5, limit=10, offset=5
        )
    )
    assert len(ir_models) == 10
    assert ir_models[0]['id'] != first_record['id']

    # default batch size and limit and offset
    ir_models = list(
        IRModel.search_read_all(
            [], None, ['rec_name'],
            limit=10, offset=5
        )
    )
    assert len(ir_models) == 10
    assert ir_models[0]['id'] != first_record['id']


def test_find_no_filter(client):
    IRModel = client.model('ir.model')
    ir_models = IRModel.find()
    assert len(ir_models) > 0
    assert ir_models[0]['id']
    assert ir_models[0]['rec_name']


def test_raises_server_error(client):
    Model = client.model('ir.model')
    with pytest.raises(ServerError):
        Model.search(1)


def test_raises_client_error():
    with pytest.raises(ClientError):
        Client('demo', 'wrong-api-key')



from requests_oauthlib import OAuth2Session


class Session(OAuth2Session):

    client_id = None
    client_secret = None

    def __init__(self, subdomain, **kwargs):
        client_id = self.client_id
        client_secret = self.client_secret
        self.fulfil_subdomain = subdomain
        if not (client_id and client_secret):
            raise Exception('Missing client_id or client_secret.')
        super(Session, self).__init__(client_id=client_id, **kwargs)

    @classmethod
    def setup(cls, client_id, client_secret):
        """Configure client in session
        """
        cls.client_id = client_id
        cls.client_secret = client_secret

    @property
    def base_url(self):
        if self.fulfil_subdomain == 'localhost':
            return 'http://localhost:8000/'
        else:
            return 'https://%s.fulfil.io/' % self.fulfil_subdomain

    def create_authorization_url(self, redirect_uri, scope, **kwargs):
        self.redirect_uri = redirect_uri
        self.scope = scope
        return self.authorization_url(
            self.base_url + 'oauth/authorize', **kwargs)

    def get_token(self, code):
        token_url = self.base_url + 'oauth/token'
        return self.fetch_token(
            token_url, client_secret=self.client_secret, code=code
        )



# -*- coding: utf-8 -*-
# pylint: disable=D102
"""
Test fulfilio data structures

pylint option block-disable
"""
import pytest
import random
from decimal import Decimal
from babel.numbers import format_currency

from money import Money
from fulfil_client.model import (
    ModificationTrackingDict, Query, StringType, MoneyType
)


@pytest.fixture
def mtd():
    """
    Return a sample Modification Tracking Dictionary
    """
    return ModificationTrackingDict({
        'a': 'apple',
        'b': 'box',
        'l': [1, 2, 3],
    })


class TestModificationTrackingDict(object):

    def test_no_changes_on_initial_dict(self, mtd):
        assert len(mtd.changes) == 0

    def test_no_changes_on_same_value(self, mtd):
        mtd['a'] = 'apple'  # nothing changes
        assert len(mtd.changes) == 0

    def test_no_changes_on_same_value_on_update(self, mtd):
        mtd.update({'a': 'apple'})
        assert len(mtd.changes) == 0

    def test_changes_on_setter(self, mtd):
        mtd['b'] = 'ball'   # big change
        assert len(mtd.changes) == 1
        assert 'b' in mtd.changes

    def test_changes_on_update(self, mtd):
        mtd.update({'b': 'ball'})   # big change
        assert len(mtd.changes) == 1
        assert 'b' in mtd.changes

    def test_changes_on_new_key(self, mtd):
        mtd['c'] = 'cat'
        assert len(mtd.changes) == 1
        assert 'c' in mtd.changes


@pytest.fixture
def query(client):
    return Query(
        client.model('res.user'),
    )


class TestQuery(object):

    def test_copyability_of_query(self, query):
        query._copy()

    def test_query_first(self, query):
        assert query.first()

    def test_query_count(self, query):
        assert query.count()

    def test_query_all(self, query):
        assert query.all()


@pytest.fixture
def res_user_model(Model):
    class ResUserModel(Model):
        __model_name__ = 'res.user'
        name = StringType()
    return ResUserModel


@pytest.fixture
def sale_order_model(Model):
    class SaleOrderModel(Model):
        __model_name__ = 'sale.sale'
        _eager_fields = set(['currency.code'])

        number = StringType()
        total_amount = MoneyType('currency_code')

        @property
        def currency_code(self):
            return self._values['currency.code']

    return SaleOrderModel


@pytest.fixture
def product_model(Model):
    class ProductModel(Model):
        __model_name__ = 'product.product'

        list_price = MoneyType('currency_code')

        @property
        def currency_code(self):
            return 'USD'

    return ProductModel


@pytest.fixture
def contact_model(Model):
    class ContactModel(Model):
        __model_name__ = 'party.party'

        name = StringType()
        credit_limit_amount = MoneyType('currency_code')

        @property
        def currency_code(self):
            return 'USD'
    return ContactModel


@pytest.fixture
def module_model(Model):
    class ModuleModel(Model):
        __model_name__ = 'ir.module'
        name = StringType()
    return ModuleModel


class TestModel(object):

    def test_model_change_tracking(self, res_user_model):
        user = res_user_model.query.first()
        user.name = user.name
        assert not bool(user.changes)

        user.name = "Not real name"
        assert 'name' in user.changes

    def test_equality_of_saved_records(self, res_user_model):
        user = res_user_model.query.first()
        user_again = res_user_model.query.get(user.id)
        assert user == user_again

    def test_inequality_of_saved_records(self, res_user_model, module_model):
        assert res_user_model.query.first() != module_model.query.first()

    def test_inequality_of_changed_models(self, res_user_model):
        user = res_user_model.query.first()
        user_again = res_user_model.query.get(user.id)
        user.display_name = "something else"
        assert user == user_again

    def test_api_url(self, res_user_model):
        user = res_user_model.query.first()
        assert user.__url__
        assert user.__client_url__


class TestMoneyType(object):

    def test_display_format(self, sale_order_model):
        order = sale_order_model.query.first()
        assert isinstance(order.total_amount, Money)
        assert isinstance(order.total_amount.amount, Decimal)
        assert order.total_amount.format('en_US') == format_currency(
            order._values['total_amount'],
            currency=order._values['currency.code'],
            locale='en_US'
        )
        assert order.total_amount.format('fr_FR') == format_currency(
            order._values['total_amount'],
            currency=order._values['currency.code'],
            locale='fr_FR'
        )

    def test_setting_values(self, product_model):
        product = product_model.query.first()

        new_price = Decimal(random.choice(list(range(1, 1000))))
        product.list_price = new_price
        product.save()

        list_price = product_model.query.first().list_price
        assert list_price.amount == new_price
        assert list_price.currency == 'USD'  # hard coded in model property

    def test_none(self, contact_model):

        contact = contact_model.query.first()

        contact.credit_limit_amount = None
        contact.save()

        credit_limit = contact.query.first().credit_limit_amount
        assert credit_limit is None

        contact.credit_limit_amount = Decimal('100000')
        contact.save()

        credit_limit = contact.query.first().credit_limit_amount
        assert credit_limit.amount == Decimal('100000')
        assert credit_limit.currency == 'USD'  # hard coded in model property



# -*- coding: UTF-8 -*-
import datetime
from decimal import Decimal
from collections import namedtuple
try:
    import simplejson as json
except ImportError:
    import json
import base64


class JSONDecoder(object):

    decoders = {}

    @classmethod
    def register(cls, klass, decoder):
        assert klass not in cls.decoders
        cls.decoders[klass] = decoder

    def __call__(self, dct):
        if dct.get('__class__') in self.decoders:
            return self.decoders[dct['__class__']](dct)
        return dct


JSONDecoder.register(
    'datetime',
    lambda dct: datetime.datetime(
        dct['year'], dct['month'], dct['day'],
        dct['hour'], dct['minute'], dct['second'], dct['microsecond']
    )
)
JSONDecoder.register(
    'date',
    lambda dct: datetime.date(dct['year'], dct['month'], dct['day'])
)
JSONDecoder.register(
    'time',
    lambda dct: datetime.time(
        dct['hour'], dct['minute'], dct['second'], dct['microsecond']
    )
)
JSONDecoder.register(
    'timedelta',
    lambda dct: datetime.timedelta(seconds=dct['seconds'])
)


def _bytes_decoder(dct):
    cast = bytearray if bytes == str else bytes
    return cast(base64.decodestring(dct['base64'].encode('utf-8')))


JSONDecoder.register('bytes', _bytes_decoder)
JSONDecoder.register(
    'Decimal', lambda dct: Decimal(dct['decimal'])
)


dummy_record = namedtuple('Record', ['model_name', 'id', 'rec_name'])
JSONDecoder.register(
    'Model', lambda dct: dummy_record(
        dct['model_name'], dct['id'], dct['rec_name']
    )
)


def parse_async_result(dct):
    from .client import AsyncResult
    return AsyncResult(dct['task_id'], dct['token'], None)


JSONDecoder.register(
    'AsyncResult', parse_async_result
)


class JSONEncoder(json.JSONEncoder):

    serializers = {}

    def __init__(self, *args, **kwargs):
        super(JSONEncoder, self).__init__(*args, **kwargs)
        # Force to use our custom decimal with simplejson
        self.use_decimal = False

    @classmethod
    def register(cls, klass, encoder):
        assert klass not in cls.serializers
        cls.serializers[klass] = encoder

    def default(self, obj):
        marshaller = self.serializers.get(
            type(obj),
            super(JSONEncoder, self).default
        )
        return marshaller(obj)


JSONEncoder.register(
    datetime.datetime,
    lambda o: {
        '__class__': 'datetime',
        'year': o.year,
        'month': o.month,
        'day': o.day,
        'hour': o.hour,
        'minute': o.minute,
        'second': o.second,
        'microsecond': o.microsecond,
        'iso_string': o.isoformat(),
    })
JSONEncoder.register(
    datetime.date,
    lambda o: {
        '__class__': 'date',
        'year': o.year,
        'month': o.month,
        'day': o.day,
        'iso_string': o.isoformat(),
    })
JSONEncoder.register(
    datetime.time,
    lambda o: {
        '__class__': 'time',
        'hour': o.hour,
        'minute': o.minute,
        'second': o.second,
        'microsecond': o.microsecond,
    })
JSONEncoder.register(
    datetime.timedelta,
    lambda o: {
        '__class__': 'timedelta',
        'seconds': o.total_seconds(),
    })
_bytes_encoder = lambda o: {
    '__class__': 'bytes',
    'base64': base64.encodestring(o).decode('utf-8'),
    }
JSONEncoder.register(bytes, _bytes_encoder)
JSONEncoder.register(bytearray, _bytes_encoder)
JSONEncoder.register(
    Decimal,
    lambda o: {
        '__class__': 'Decimal',
        'decimal': str(o),
    })



# -*- coding: utf-8 -*-
"""
Fulfil.IO Model Helper

A collection of model layer APIs to write lesser code
and better
"""
import six
import logging
import functools
import warnings
from datetime import datetime, date
from copy import copy
from decimal import Decimal
from money import Money
from itertools import izip

import fulfil_client
from fulfil_client.client import loads, dumps

cache_logger = logging.getLogger('fulfil_client.cache')


class BaseType(object):
    """
    A django field like object that implements a descriptor.
    """

    # Eager load this field
    eager = True

    def __init__(self, cast, required=False, default=None):
        # this will be auto discovered by a meta class
        self.name = None

        self.cast = cast
        self.default = default
        self.required = required

    def __get__(self, instance, owner):
        if instance:
            return instance._values.get(self.name, self.default)
        else:
            return self

    def convert(self, value):
        if value is None:
            return
        if isinstance(value, self.cast):
            return value
        return self.cast(value)

    def __set__(self, instance, value):
        instance._values[self.name] = self.convert(value)

    def __delete__(self, instance):
        del instance._values[self.name]


class IntType(BaseType):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('cast', int)
        super(IntType, self).__init__(*args, **kwargs)


class BooleanType(BaseType):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('cast', bool)
        super(BooleanType, self).__init__(*args, **kwargs)


class StringType(BaseType):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('cast', six.text_type)
        super(StringType, self).__init__(*args, **kwargs)


class DecimalType(BaseType):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('cast', Decimal)
        super(DecimalType, self).__init__(*args, **kwargs)


class FloatType(BaseType):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('cast', float)
        super(FloatType, self).__init__(*args, **kwargs)


class DateTime(BaseType):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('cast', datetime)
        super(DateTime, self).__init__(*args, **kwargs)


class Date(BaseType):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('cast', date)
        super(Date, self).__init__(*args, **kwargs)


class One2ManyType(BaseType):

    def __init__(self, model_name, cache=False, *args, **kwargs):
        self.model_name = model_name
        self.cache = cache
        kwargs.setdefault('cast', list)
        super(One2ManyType, self).__init__(*args, **kwargs)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        if instance._values.get(self.name):
            model = instance.__modelregistry__[self.model_name]
            if self.cache:
                return model.from_cache(instance._values.get(self.name))
            return model.from_ids(instance._values.get(self.name))
        return instance._values.get(self.name)


class MoneyType(DecimalType):
    """
    Built on top of the decimal field, but also understands the
    currency with which the amount is defined.

    Usage examples:

        sale.total_amount.amount

    Formatting for web

        sale.total_amount.format()

    :param currency_field: Name of the field that will have the 3 char
                           currency code
    """

    def __init__(self, currency_field, *args, **kwargs):
        self.currency_field = currency_field
        super(MoneyType, self).__init__(*args, **kwargs)

    def __get__(self, instance, owner):
        if instance:
            if instance._values.get(self.name, self.default) is None:
                return None
            return Money(
                instance._values.get(self.name, self.default),
                getattr(instance, self.currency_field)
            )
        else:
            return self


class ModelType(IntType):
    """
    Builds a field that represents a Foreign Key relationship to another
    model.

    :param model_name: Name of the model to which this model has a FK
    :param cache: If set, it looks up the record in the cache backend of the
                  underlying model before querying the server to fetch records.
    """
    def __init__(self, model_name, cache=False, *args, **kwargs):
        self.model_name = model_name
        self.cache = cache
        super(ModelType, self).__init__(*args, **kwargs)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        if instance._values.get(self.name):
            model = instance.__modelregistry__[self.model_name]
            if self.cache:
                return model.from_cache(instance._values.get(self.name))
            return model.get_by_id(instance._values.get(self.name))
        return instance._values.get(self.name)


class NamedDescriptorResolverMetaClass(type):
    """
    A metaclass that discovers field names
    """

    def __new__(cls, classname, bases, class_dict):
        abstract = class_dict.get('__abstract__', False)
        model_name = class_dict.get('__model_name__')

        if not abstract and not model_name:
            for base in bases:
                if hasattr(base, '__model_name__'):
                    model_name = base.__model_name__
                    break
            else:
                raise Exception('__model_name__ not defined for model')

        fields = set([])
        eager_fields = set([])
        for base in bases:
            if hasattr(base, '_fields'):
                fields |= set(base._fields)
            if hasattr(base, '_eager_fields'):
                eager_fields |= set(base._eager_fields)

        fields |= class_dict.get('_fields', set([]))
        eager_fields |= class_dict.get('_eager_fields', set([]))

        # Iterate through the new class' __dict__ to:
        #
        # * update all recognised NamedDescriptor member names
        # * find lazy loaded fields
        # * find eager_loaded fields
        for name, attr in class_dict.items():
            if isinstance(attr, BaseType):
                attr.name = name
                fields.add(name)
                if attr.eager:
                    eager_fields.add(name)

        class_dict['_eager_fields'] = eager_fields
        class_dict['_fields'] = fields | eager_fields

        # Call super and continue class creation
        rv = type.__new__(cls, classname, bases, class_dict)
        if not abstract:
            rv.__modelregistry__[model_name] = rv
        return rv


class ModificationTrackingDict(dict):
    """
    A change tracking dictionary
    """

    def __init__(self, *args, **kwargs):
        self.changes = set([])
        super(ModificationTrackingDict, self).__init__(*args, **kwargs)

    def __setitem__(self, key, val):
        if key not in self or self[key] != val:
            self.changes.add(key)
        dict.__setitem__(self, key, val)

    def update(self, *args, **kwargs):
        """
        Update does not call __setitem__ by default
        """
        for k, v in dict(*args, **kwargs).items():
            self[k] = v


def return_instances(function):
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        query = args[0]
        map_fn = query.instance_class
        for record in function(*args, **kwargs):
            yield map_fn(record) if map_fn else record
    return wrapper


def return_instance(function):
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        query = args[0]
        result = function(*args, **kwargs)
        if result is None:
            return None
        if query.instance_class:
            return query.instance_class(**result)
        else:
            return result
    return wrapper


class classproperty(object):    # NOQA
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


class Query(object):
    """
    A sqlalchemy like query object giving developers familiar primitives.

    The method are limited to what are reasonable over an API.
    """

    def __init__(self, model, instance_class=None):
        self.rpc_model = model
        self.instance_class = instance_class or None
        self.domain = []
        self._limit = None
        self._offset = None
        self._order_by = None
        self.active_only = True

    @property
    def fields(self):
        return self.instance_class and tuple(self.instance_class._fields) or \
                None

    def __copy__(self):
        """
        Change the copy behavior of query.

        Maintain references to model and instance class while building a new
        list for domain and order by
        """
        newone = type(self)(self.rpc_model, self.instance_class)

        # immutable types
        newone._limit = self._limit
        newone._offset = self._offset
        newone.active_only = self.active_only

        # Duplicate the lists
        if self._order_by:
            newone._order_by = self._order_by[:]
        newone.domain = self.domain[:]
        return newone

    @property
    def context(self):
        "Return the context to execute the query"
        return {
            'active_test': self.active_only,
        }

    def _copy(self):
        "Internal method to make copies of the query"
        return copy(self)

    @return_instances
    def all(self):
        """
        Return the results represented by this Query as a list.

        .. versionchanged:: 0.10.0

            Returns an iterator that lazily loads
            records instead of fetching thousands
            of records at once.
        """
        return self.rpc_model.search_read_all(
            self.domain,
            self._order_by,
            self.fields,
            context=self.context,
            offset=self._offset or 0,
            limit=self._limit,
        )

    def count(self):
        "Return a count of rows this Query would return."
        return self.rpc_model.search_count(
            self.domain, context=self.context
        )

    def exists(self):
        """
        A convenience method that returns True if a record
        satisfying the query exists
        """
        return self.rpc_model.search_count(
            self.domain, context=self.context
        ) > 0

    def show_active_only(self, state):
        """
        Set active only to true or false on a copy of this query
        """
        query = self._copy()
        query.active_only = state
        return query

    def filter_by(self, **kwargs):
        """
        Apply the given filtering criterion to a copy of this Query, using
        keyword expressions.
        """
        query = self._copy()
        for field, value in kwargs.items():
            query.domain.append(
                (field, '=', value)
            )
        return query

    def filter_by_domain(self, domain):
        """
        Apply the given domain to a copy of this query
        """
        query = self._copy()
        query.domain = domain
        return query

    @return_instance
    def first(self):
        """
        Return the first result of this Query or None if the result
        doesn't contain any row.
        """
        results = self.rpc_model.search_read(
            self.domain, None, 1, self._order_by, self.fields,
            context=self.context
        )
        return results and results[0] or None

    @return_instance
    def get(self, id):
        """
        Return an instance based on the given primary key identifier,
        or None if not found.

        This returns a record whether active or not.
        """
        ctx = self.context.copy()
        ctx['active_test'] = False
        results = self.rpc_model.search_read(
            [('id', '=', id)],
            None, None, None, self.fields,
            context=ctx
        )
        return results and results[0] or None

    def limit(self, limit):
        """
        Apply a LIMIT to the query and return the newly resulting Query.
        """
        query = self._copy()
        query._limit = limit
        return query

    def offset(self, offset):
        """
        Apply an OFFSET to the query and return the newly resulting Query.
        """
        query = self._copy()
        query._offset = offset
        return query

    @return_instance
    def one(self):
        """
        Return exactly one result or raise an exception.

        Raises fulfil_client.exc.NoResultFound if the query selects no rows.
        Raises fulfil_client.exc.MultipleResultsFound if multiple rows are
        found.
        """
        results = self.rpc_model.search_read(
            self.domain, 2, None, self._order_by, self.fields,
            context=self.context
        )
        if not results:
            raise fulfil_client.exc.NoResultFound
        if len(results) > 1:
            raise fulfil_client.exc.MultipleResultsFound
        return results[0]

    def order_by(self, *criterion):
        """
        apply one or more ORDER BY criterion to the query and
        return the newly resulting Query

        All existing ORDER BY settings can be suppressed by passing None -
        this will suppress any ORDER BY configured on mappers as well.
        """
        query = self._copy()
        query._order_by = criterion
        return query

    def delete(self):
        """
        Delete all records matching the query.

        Warning: This is a desctructive operation.

        Not every model allows deletion of records and several models
        even restrict based on status. For example, deleting products
        that have been transacted is restricted. Another example is sales
        orders which can be deleted only when they are draft.

        If deletion fails, a server error is thrown.
        """
        ids = self.rpc_model.search(self.domain, context=self.context)
        if ids:
            self.rpc_model.delete(ids)

    def archive(self):
        """
        Archives (soft delete) all the records matching the query.

        This assumes that the model allows archiving (not many do - especially
        transactional documents).

        Internal implementation sets the active field to False.
        """
        ids = self.rpc_model.search(self.domain, context=self.context)
        if ids:
            self.rpc_model.write(ids, {'active': False})


@six.add_metaclass(NamedDescriptorResolverMetaClass)
class Model(object):
    """
    Active record design pattern for RPC models.
    """

    __abstract__ = True

    fulfil_client = None
    cache_backend = None
    cache_expire = None

    id = IntType()

    def __init__(self, values=None, id=None, **kwargs):
        values = values or {}
        values.update(kwargs)

        if id is not None:
            values['id'] = id

        # Now create a modification tracking dictionary
        self._values = ModificationTrackingDict(values)

    @classmethod
    def get_cache_key(cls, id):
        "Return a cache key for the given id"
        return '%s:%s:%s' % (
            cls.fulfil_client.subdomain,
            cls.__model_name__,
            id
        )

    @property
    def cache_key(self):
        return self.get_cache_key(self.id)

    @classmethod
    def from_cache_multi(cls, ids, ignore_misses=False):
        """
        Check if a record is in cache. If it is, load from there, if not
        load the record and then cache it, but in bulk.

        For performance, you can opt to ignore fetching records that are
        not already on the cache. While this can give you a performance
        boost, this may result in some or more records missing. Use with
        care, you've been warned!

        :param ignore_misses: If True, then the returned set will not
                              include records that are not already in
                              cache.
        """
        results = []
        misses = []
        if not cls.cache_backend:
            misses = ids
        else:
            cached_values = cls.cache_backend.mget(map(cls.get_cache_key, ids))
            for id, cached_value in izip(ids, cached_values):
                if cached_value:
                    results.append(cls(id=id, values=loads(cached_value)))
                else:
                    misses.append(id)

        if misses:
            cache_logger.warn(
                "MISS::MULTI::%s::%s" % (cls.__model_name__, misses)
            )

        if misses and not ignore_misses:
            # Get the records in bulk for misses
            rows = cls.rpc.read(misses, tuple(cls._fields))
            for row in rows:
                record = cls(id=row['id'], values=row)
                record.store_in_cache()
                results.append(record)

        return sorted(results, key=lambda r: ids.index(r.id))

    @classmethod
    def from_cache(cls, id):
        """
        Check if a record is in cache. If it is load from there, if not
        load the record and then cache it.
        """
        if isinstance(id, (list, tuple)):
            message = "For list and ids use from_cache_multi"
            warnings.warn(message, DeprecationWarning, stacklevel=2)
            return cls.from_cache_multi(id)

        key = cls.get_cache_key(id)
        cached_value = cls.cache_backend and cls.cache_backend.get(key)

        if cached_value:
            cache_logger.debug("HIT::%s" % key)
            return cls(id=id, values=loads(cached_value))

        cache_logger.warn("MISS::%s" % key)
        record = cls(id=id)
        record.refresh()
        record.store_in_cache()
        return record

    def store_in_cache(self):
        """
        Save the values to a cache.
        """
        if self.cache_backend:
            self.cache_backend.set(
                self.cache_key, dumps(self._values), self.cache_expire
            )

    def invalidate_cache(self):
        if self.cache_backend:
            self.cache_backend.delete(self.cache_key)

    @classmethod
    def from_ids(cls, ids):
        """
        Create multiple active resources at once
        """
        return list(map(cls, cls.rpc.read(ids, tuple(cls._eager_fields))))

    @property
    def changes(self):
        """
        Return a set of changes
        """
        return dict([
            (field_name, self._values[field_name])
            for field_name in self._values.changes
        ])

    @classproperty
    def query(cls):     # NOQA
        return Query(cls.get_rpc_model(), cls)

    @property
    def has_changed(self):
        "Return True if the record has changed"
        return len(self._values) > 0

    @classproperty
    def rpc(cls):       # NOQA
        "Returns an RPC client for the Fulfil.IO model with same name"
        return cls.get_rpc_model()

    @classmethod
    def get_rpc_model(cls):
        "Returns an instance of the model record"
        return cls.fulfil_client.model(cls.__model_name__)

    @classmethod
    def get_rpc_record(cls, id):
        "Returns an instance for Fulfil.IO Client Record"
        return cls.fulfil_client.record(cls.__model_name__, id)

    @classmethod
    def get_by_id(cls, id):
        "Given an integer ID, fetch the record from fulfil.io"
        return cls(values=cls.rpc.read([id], tuple(cls._eager_fields))[0])

    def delete(self):
        """
        Delete the record
        """
        return self.rpc.delete([self.id])

    def refresh(self):
        """
        Refresh a record by fetching again from the API.
        This also resets the modifications in the record.
        """
        self.invalidate_cache()

        assert self.id, "Cannot refresh unsaved record"
        self._values = ModificationTrackingDict(
            self.rpc.read([self.id], tuple(self._fields))[0]
        )

    def save(self):
        "Save as a new record if there is no id, or update record with id"
        if self.id:
            if self.changes:
                self.rpc.write([self.id], self.changes)
        else:
            self.id = self.rpc.create([self._values])[0]

        # Either way refresh the record after saving
        self.refresh()

        return self

    def __eq__(self, other):
        if other is None:
            return False
        if other.__model_name__ != self.__model_name__:
            # has to be of the same model
            return False
        if self.id and (other.id != self.id):
            # If the record has an ID the other one should
            # have the same.
            return False
        if not self.id and (self._values != other._values):
            # Unsaved records are same only if _values
            # of both are the same.
            return False
        return True

    @property
    def __url__(self):
        "Return the API URL for the record"
        return '/'.join([
            self.rpc.client.base_url,
            self.__model_name__,
            six.text_type(self.id)
        ])

    @property
    def __client_url__(self):
        "Return the Client URL for the record"
        return '/'.join([
            self.rpc.client.host,
            'client/#/model',
            self.__model_name__,
            six.text_type(self.id)
        ])


def model_base(fulfil_client, cache_backend=None, cache_expire=10 * 60):
    """
    Return a Base Model class that binds to the fulfil client instance and
    the cache instance.

    This design is inspired by the declarative base pattern in SQL Alchemy.
    """
    return type(
        'BaseModel',
        (Model,),
        {
            'fulfil_client': fulfil_client,
            'cache_backend': cache_backend,
            'cache_expire': cache_expire,
            '__abstract__': True,
            '__modelregistry__': {},
        },
    )



import base64
import json
import logging
from datetime import datetime
from functools import partial, wraps

import requests
from .serialization import JSONDecoder, JSONEncoder


request_logger = logging.getLogger('fulfil_client.request')
dumps = partial(json.dumps, cls=JSONEncoder)
loads = partial(json.loads, object_hook=JSONDecoder())


class Error(Exception):
    def __init__(self, message, code):
        super(Exception, self).__init__(message)
        self.code = code


class ServerError(Error):
    pass


class ClientError(Error):
    pass


def json_response(function):
    @wraps(function)
    def wrapper(*args, **kwargs):
        rv = function(*args, **kwargs)
        if rv.status_code != requests.codes.ok:
            if 400 <= rv.status_code and rv.status_code < 500:
                raise ClientError(loads(rv.text), rv.status_code)
            raise ServerError(rv.text, rv.status_code)
        return loads(rv.text)
    return wrapper


class SessionAuth(requests.auth.AuthBase):
    "Session Authentication"
    type_ = 'Session'

    def __init__(self, login, user_id, session):
        self.login = login
        self.user_id = user_id
        self.session = session

    def __call__(self, r):
        r.headers['Authorization'] = 'Session ' + base64.b64encode(
            '%s:%s:%s' % (self.login, self.user_id, self.session)
        )
        return r


class BearerAuth(requests.auth.AuthBase):
    "Bearer Authentication"
    type_ = 'BearerAuth'

    def __init__(self, access_token):
        self.access_token = access_token

    def __call__(self, r):
        r.headers['Authorization'] = 'Bearer ' + self.access_token
        return r


class APIKeyAuth(requests.auth.AuthBase):
    "API key based Authentication"
    type_ = 'APIKey'

    def __init__(self, api_key):
        self.api_key = api_key

    def __call__(self, r):
        r.headers['x-api-key'] = self.api_key
        return r


class Client(object):

    def __init__(self, subdomain,
                 api_key=None, context=None, auth=None,
                 user_agent="Python Client"):
        self.subdomain = subdomain

        if self.subdomain == 'localhost':
            self.host = 'http://localhost:8000'
        else:
            self.host = 'https://%s.fulfil.io' % self.subdomain

        self.base_url = '%s/api/v1' % self.host

        self.session = requests.Session()
        if api_key is not None:
            self.set_auth(APIKeyAuth(api_key))
        else:
            self.set_auth(auth)

        self.session.headers.update({
            'User-Agent': user_agent,
        })

        self.context = {}
        if context is not None:
            self.context.update(context)
        if context is None and self.session.auth:
            # context is not defined, but auth is there.
            # try and get a context
            self.refresh_context()

    def set_auth(self, auth):
        self.session.auth = auth
        if auth is None:
            return
        if isinstance(auth, BearerAuth):
            self.base_url = '%s/api/v2' % self.host

    def set_user_agent(self, user_agent):
        self.session.headers.update({
            'User-Agent': user_agent
        })

    def refresh_context(self):
        """
        Get the default context of the user and save it
        """
        User = self.model('res.user')

        self.context = User.get_preferences(True)
        return self.context

    def today(self):
        Date = self.model('ir.date')
        rv = Date.today()
        return rv

    def model(self, name):
        return Model(self, name)

    def record(self, model_name, id):
        return Record(self.model(model_name), id)

    def report(self, name):
        return Report(self, name)

    def interactive_report(self, name):
        return InteractiveReport(self, name)

    def login(self, login, password, set_auth=False):
        """
        Attempts a login to the remote server
        and on success returns user id and session
        or None

        Warning: Do not depend on this. This will be deprecated
        with SSO.

        param set_auth: sets the authentication on the client
        """
        rv = self.session.post(
            self.host,
            dumps({
                "method": "common.db.login",
                "params": [login, password]
            }),
        )
        rv = loads(rv.content)['result']
        if set_auth:
            self.set_auth(
                SessionAuth(login, *rv)
            )
        return rv

    def is_auth_alive(self):
        "Return true if the auth is not expired, else false"
        model = self.model('ir.model')
        try:
            model.search([], None, 1, None)
        except ClientError as err:
            if err and err.message['code'] == 403:
                return False
            raise
        except Exception:
            raise
        else:
            return True


class Record(object):
    def __init__(self, model, id):
        self.model = model
        self.id = id

    def __getattr__(self, name):
        return getattr(self.model, name)

    def update(self, data=None, **kwargs):
        """
        Update the record right away.

        :param data: dictionary of changes
        :param kwargs: possibly a list of keyword args to change
        """
        if data is None:
            data = {}
        data.update(kwargs)
        return self.model.write([self.id], data)


class Model(object):

    def __init__(self, client, model_name):
        self.client = client
        self.model_name = model_name

    def __getattr__(self, name):
        @json_response
        def proxy_method(*args, **kwargs):
            context = self.client.context.copy()
            context.update(kwargs.pop('context', {}))
            request_logger.debug(
                "%s.%s::%s::%s" % (
                    self.model_name, name, args, kwargs
                )
            )
            return self.client.session.put(
                self.path + '/%s' % name,
                dumps(args),
                params={
                    'context': dumps(context),
                }
            )
        return proxy_method

    @property
    def path(self):
        return '%s/model/%s' % (self.client.base_url, self.model_name)

    @json_response
    def get(self, id, context=None):
        ctx = self.client.context.copy()
        ctx.update(context or {})
        return self.client.session.get(
            self.path + '/%d' % id,
            params={
                'context': dumps(ctx),
            }
        )

    def search_read_all(self, domain, order, fields, batch_size=500,
                        context=None, offset=0, limit=None):
        """
        An endless iterator that iterates over records.

        :param domain: A search domain
        :param order: The order clause for search read
        :param fields: The fields argument for search_read
        :param batch_size: The optimal batch size when sending paginated
                           requests
        """
        if context is None:
            context = {}

        if limit is None:
            # When no limit is specified, all the records
            # should be fetched.
            record_count = self.search_count(domain, context=context)
            end = record_count + offset
        else:
            end = limit + offset

        for page_offset in range(offset, end, batch_size):
            if page_offset + batch_size > end:
                batch_size = end - page_offset
            for record in self.search_read(
                    domain, page_offset, batch_size,
                    order, fields, context=context):
                yield record

    @json_response
    def find(self, filter=None, page=1, per_page=10, fields=None, context=None):
        """
        Find records that match the filter.

        Pro Tip: The fields could have nested fields names if the field is
        a relationship type. For example if you were looking up an order
        and also want to get the shipping address country then fields would be:

            `['shipment_address', 'shipment_address.country']`

        but country in this case is the ID of the country which is not very
        useful if you don't already have a map. You can fetch the country code
        by adding `'shipment_address.country.code'` to the fields.

        :param filter: A domain expression (Refer docs for domain syntax)
        :param page: The page to fetch to get paginated results
        :param per_page: The number of records to fetch per page
        :param fields: A list of field names to fetch.
        :param context: Any overrides to the context.
        """
        if filter is None:
            filter = []
        return self.client.session.get(
            self.path,
            params={
                'filter': dumps(filter or []),
                'page': page,
                'per_page': per_page,
                'field': fields,
                'context': dumps(context or self.client.context),
            }
        )


class Report(object):

    def __init__(self, client, report_name):
        self.client = client
        self.report_name = report_name

    @property
    def path(self):
        return '%s/report/%s' % (self.client.base_url, self.report_name)

    @json_response
    def execute(self, records=None, data=None, **kwargs):
        context = self.client.context.copy()
        context.update(kwargs.pop('context', {}))
        return self.client.session.put(
            self.path,
            json={
                'objects': records or [],
                'data': data or {},
            },
            params={
                'context': dumps(context),
            }
        )


class InteractiveReport(object):

    def __init__(self, client, model_name):
        self.client = client
        self.model_name = model_name

    @property
    def path(self):
        return '%s/model/%s/execute' % (
            self.client.base_url, self.model_name
        )

    @json_response
    def execute(self, **kwargs):
        context = self.client.context.copy()
        context.update(kwargs.pop('context', {}))
        return self.client.session.put(
            self.path,
            dumps([kwargs]),
            params={
                'context': dumps(context),
            }
        )


class AsyncResult(object):
    """
    When the server returns an AsyncResult (usually for long running tasks),
    an instance of this class is created in the response. The object provides
    a convenient wrapper with which you can poll for the status of the task
    and result.
    """

    PENDING = 'PENDING'
    STARTED = 'STARTED'
    FAILURE = 'FAILURE'
    SUCCESS = 'SUCCESS'
    RETRY = 'RETRY'

    def __init__(self, task_id, token, client):
        self.task_id = task_id
        self.token = token
        self.client = client
        self.result = None
        self.state = self.PENDING
        self.start_time = datetime.utcnow()

    @property
    def path(self):
        return '%s/async-result' % (self.client.base_url)

    def bind(self, client):
        self.client = client
        return self

    @json_response
    def _fetch_result(self):
        if self.client is None:
            raise Exception(
                "Unbound AsyncResults cannot refresh status.\n"
                "Hint: Bind `result.bind(client)` and try again."
            )
        return self.client.session.post(
            self.path,
            json={
                'tasks': [
                    [self.task_id, self.token]
                ]
            },
        )

    def refresh_if_needed(self):
        """
        Refresh the status of the task from server if required.
        """
        if self.state in (self.PENDING, self.STARTED):
            try:
                response, = self._fetch_result()['tasks']
            except (KeyError, ValueError):
                raise Exception(
                    "Unable to find results for task."
                )

            if 'error' in response:
                self.state == self.FAILURE
                raise ServerError(response['error'])

            if 'state' in response:
                self.state = response['state']
                self.result = response['result']

    def failed(self):
        """
        Returns true of the task failed
        """
        self.refresh_if_needed()
        return self.state == self.FAILURE

    def ready(self):
        """
        Returns True if the task has been executed.

        If the task is still running, pending, or is waiting for retry then
        False is returned.
        """
        self.refresh_if_needed()
        return self.state == self.SUCCESS

    def wait(self, timeout=None):
        """
        Wait until task is ready, and return its result.

        Not implemented yet
        """
        raise Exception("Not implemented yet")

    @property
    def time_lapsed(self):
        """
        Time lapsed since the task was started

        Returned only when the task is still in progress
        """
        return datetime.utcnow() - self.start_time



# -*- coding: utf-8 -*-

__author__ = 'Fulfil.IO Inc.'
__email__ = 'hello@fulfil.io'
__version__ = '0.11.0'

# flake8: noqa

from .client import Client, Model, ServerError, SessionAuth, APIKeyAuth, \
    BearerAuth, ClientError
